/*************************************************************************
  
  no_block.c
	 
     Provides a command which simulates a non-blocking tclsh.
			
  Please direct any questions or comments to:
			  
     Brian Starr
     bstarr@ics.uci.edu
							
  This code was created as part of the Cafe ConstructionKit project at
  the University of California, Irvine.  For more information on the
  Cafe ConstructionKit project, contact:
									
     Mark Ackerman
     vl@ics.uci.edu

  Release date:  v1.0 5/20/95

  Copyright (c) 1995 The Regents of the University of California.
  All rights reserved.
  
  Permission is hereby granted, without written agreement and without
  license or royalty fees, to use, copy, modify, and distribute this
  software and its documentation for any purpose, provided that the
  above copyright notice and the following two paragraphs appear in
  all copies of this software.
  
  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  
  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
  ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
  PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ************************************************************************/

#include <stdio.h>
#include <tcl.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include "no_block.h"
#include <malloc.h>

extern "C"	int     isatty _ANSI_ARGS_((int fd));
extern "C" int     select(int,fd_set*,fd_set*,fd_set*,struct timeval *);
extern "C" void     bzero(char*,int);

/* get_user_input - print a prompt if need be, and try to read from input file*/
static int get_user_input(tcl_input *inp, Tcl_Interp* ti, int argc, 
	char* argv[])
{
	char buffer[1000];
	int  do_prompt = 0;
	struct timeval time = {0,0};
	fd_set read,write,error;

	if (argc>1)
		Tcl_GetInt(ti,argv[1],&do_prompt);
	do_prompt |= inp->needs_prompt;
	if ((do_prompt) && (inp->tty)) {
		char *promptCmd;

		promptCmd = Tcl_GetVar(ti,
			inp->gotPartial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
		if (promptCmd == NULL) {
			defaultPrompt:
			if (!inp->gotPartial)
				fputs("% ", stdout);
		} else {
			int code = Tcl_Eval(ti, promptCmd);
			if (code != TCL_OK) {
				Tcl_AddErrorInfo(ti,"\n    (script that generates prompt)");
				fprintf(stderr, "%s\n", ti->result);
				goto defaultPrompt;
			}
		}
      fflush(stdout);
		inp->needs_prompt = False;
	}

	clearerr(inp->file);

	FD_ZERO(&read);
	FD_ZERO(&write);
	FD_ZERO(&error);
	FD_SET((int)inp->file->_file,&read);
	FD_SET((int)inp->file->_file,&error);

	ti->result = "1";
	if (select(1,&read,&write,&error,&time)) {
		buffer[0] = 0;
		inp->needs_prompt = True;
		if (fgets(buffer, 1000, inp->file) == NULL) {
			if (ferror(inp->file)) {
				if (errno == EINTR) {
					if (tcl_AsyncReady) {
						(void) Tcl_AsyncInvoke((Tcl_Interp *) NULL, 0);
					}
					clearerr(inp->file);
					return TCL_OK;
				} else {
				/* got input error */
					return TCL_ERROR;
				}
			} else if (!inp->gotPartial) {
				/* Got end of file */
				ti->result = "0";
				return TCL_OK;
			}
      } else {
			char* cmd = Tcl_DStringAppend(&inp->command, buffer, -1);
			int code;
			if ((buffer[0] != 0) && !Tcl_CommandComplete(cmd)) {
				inp->gotPartial = True;
				return TCL_OK;
			}
			inp->gotPartial = False;
			code = Tcl_RecordAndEval(ti, cmd, 0);
			Tcl_DStringFree(&inp->command);
			if (code != TCL_OK) {
				fprintf(stderr, "%s\n", ti->result);
			} else if (inp->tty && (*ti->result != 0)) {
				printf("%s\n", ti->result);
			}
			Tcl_ResetResult(ti);
			ti->result = "1";
		}
	}

	return TCL_OK;
}

static void do_error(Tcl_Interp* ti, char* error)
{
	Tcl_AppendResult("Error creating new input stream.\n",error,(char*)NULL);
}

char e_bad_params[] = "Calling signiture: no_block name\n";
char e_no_mem[] = "Out of memory executing \"no_block\".\n";

int no_block_CreateInput(Tcl_Interp* ti, char* name, FILE* file)
{
	tcl_input* inp = (tcl_input*) malloc(sizeof(tcl_input));
	if (!inp) {
		do_error(ti,e_no_mem);
		return TCL_ERROR;
	}

	inp->file = file;
	inp->gotPartial = False;
	Tcl_DStringInit(&inp->command);
	inp->tty = isatty(file->_file);
	inp->needs_prompt = True;

	Tcl_CreateCommand(ti,name,get_user_input,inp, free);

	// return the name of the object added.
	Tcl_SetResult(ti,name,TCL_VOLATILE);

	return TCL_OK;

}

static int new_user_input(ClientData*x, Tcl_Interp* ti, int argc, char* argv[])
{

	if (argc!=2) {

		do_error(ti,e_bad_params);
		return TCL_ERROR;

	}

	return no_block_CreateInput(ti,argv[1],stdin);

}

int no_block_Init(Tcl_Interp* ti)
{
	Tcl_CreateCommand(ti, "no_block", new_user_input,
		(ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
	return TCL_OK;
}
